/***************************************************************************
                          quickview.cpp  -  description
                             -------------------
    begin                : Sun Jul 14 2002
    copyright            : (C) 2002 by Chris Wallace
    email                : chris.wallace@netflash.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "quickview.h"

#include <qlayout.h>
#include <qtoolbutton.h>
#include <qscrollview.h>
#include <qtextstream.h>
#include <qdir.h>
#include <qfile.h>
#include <qregexp.h>
#include <qtooltip.h>

#include <kurl.h>
#include <kurldrag.h>
#include <kglobal.h>
#include <klocale.h>
#include <kconfig.h>
#include <kprocess.h>
#include <kpopupmenu.h>
#include <kwordwrap.h>
#include <kiconloader.h>

#include <iostream>

#include "sitehandler.h"
#include "schemetree.h"
#include "schemeparser.h"
#include "rssparser.h"

#include "quickviewscroller.h"
#include "quickviewurl.h"

const unsigned int MAX_TOOLTIP_CHARS = 80;

QuickView::QuickView(QWidget *parent, const char *name )
 : QWidget(parent,name, Qt::WStyle_Customize | Qt::WStyle_DialogBorder | Qt::WStyle_Tool | Qt::WStyle_StaysOnTop)
{
  QHBoxLayout * mLay = new QHBoxLayout(this, 0, -1, "mLay");

  mFeedBack = new QToolButton(this, "mFeedBack");
  mFeedBack->setPixmap(KGlobal::iconLoader()->loadIcon("2leftarrow", KIcon::Small));
  QToolTip::add(mFeedBack, i18n("Go to the previous feed"));
  mFeedNext = new QToolButton(this, "mFeedBack");
  mFeedNext->setPixmap(KGlobal::iconLoader()->loadIcon("2rightarrow", KIcon::Small));
  QToolTip::add(mFeedNext, i18n("Go to the next feed"));

  mItemBack = new QToolButton(this, "mItemBack");
  mItemBack->setPixmap(KGlobal::iconLoader()->loadIcon("1leftarrow", KIcon::Small));
  QToolTip::add(mItemBack, i18n("Go to the previous news item"));
  mItemNext = new QToolButton(this, "mItemNext");
  mItemNext->setPixmap(KGlobal::iconLoader()->loadIcon("1rightarrow", KIcon::Small));
  QToolTip::add(mItemNext, i18n("Go to the next news item"));

  mScroller = new QuickViewScroller(this, "mScroller");
  mScroller->enableClipper(true);

  mItem = new QuickViewURL(QString("http://krss.sf.net"), QString("K.R.S.S."), mScroller->viewport(), "mItem");
  mItem->setAlignment(Qt::AlignCenter);

  mScroller->addChild(mItem, 0, 0);

  mLay->addWidget(mFeedBack);
  mLay->addWidget(mItemBack);
  mLay->addWidget(mScroller);
  mLay->addWidget(mItemNext);
  mLay->addWidget(mFeedNext);

  mScroller->setSizePolicy(QSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding));
  
  mCurrentSite = 0;
  mCurrentItem = 0;

  connect(mFeedBack, SIGNAL(clicked()), this, SLOT(slotFeedBackClicked()));
  connect(mFeedNext, SIGNAL(clicked()), this, SLOT(slotFeedNextClicked()));
  connect(mItemBack, SIGNAL(clicked()), this, SLOT(slotItemBackClicked()));
  connect(mItemNext, SIGNAL(clicked()), this, SLOT(slotItemNextClicked()));

  connect(&mTimer, SIGNAL(timeout()), this, SLOT(slotTimeout()));

  connect(mItem, SIGNAL(leftClickedURL(const QString &)), this, SLOT(slotLeftClickURL(const QString &)));
  connect(mItem, SIGNAL(rightClickedURL(const QString &)), this, SLOT(slotRightClickURL(const QString &)));

  connect(mScroller, SIGNAL(signalResized(int, int, int, int)), this, SLOT(slotScrollerResized(int, int, int, int)));

  connect(mScroller, SIGNAL(signalContentsDragEntered(QDragEnterEvent *)), this, SLOT(slotContentsDragEntered(QDragEnterEvent *)));
  connect(mScroller, SIGNAL(signalContentsDroped(QDropEvent *)), this, SLOT(slotContentsDroped(QDropEvent *)));
  connect(mItem, SIGNAL(signalDragEntered(QDragEnterEvent *)), this, SLOT(slotContentsDragEntered(QDragEnterEvent *)));
  connect(mItem, SIGNAL(signalDroped(QDropEvent *)), this, SLOT(slotContentsDroped(QDropEvent *)));

  connect(this, SIGNAL(signalDnDFinished()), this, SLOT(slotDnDFinished()));

  mDiff = 0;
  mTotal = 0;

  mSiteHandler = 0;

  //Turn on Drag 'n Drop
  mScroller->viewport()->setAcceptDrops(true);
}

QuickView::~QuickView()
{
}

void QuickView::show(void)
{
  KConfig *config = KGlobal::config();
  QString group = config->group();
  config->setGroup(QString("KRSS Quick View"));
  mScrollMethod = config->readNumEntry(QString("Scroll Direction"), (int)Horizontal);
  QSize size = config->readSizeEntry(QString("Geometry"));
  if(!size.isEmpty())
    resize(size);

  mFade = config->readBoolEntry(QString("Fade Items"), true);
  mFadeSpeed = config->readNumEntry(QString("Fade Speed"), 4);

  mItem->setUseTips(config->readBoolEntry(QString("Display Full Info"), true));

  config->setGroup(group);

  //cout << "Scroll Direction: " << mScrollMethod << endl;
  //cout << "Horizontal Direction: " << (int)Horizontal << endl;

  mTimer.start(100);

  QWidget::show();
}

void QuickView::hide(void)
{
  mTimer.stop();
  KConfig *config = KGlobal::config();
  QString group = config->group();
  config->setGroup(QString("KRSS Quick View"));
  config->writeEntry(QString("Geometry"), size());
  config->setGroup(group);

  QWidget::hide();
}

void QuickView::slotFeedBackClicked(void)
{
  goTo(PreviousSite, LastItem);
}

void QuickView::slotItemBackClicked(void)
{
  goTo(SameSite, PreviousItem);
}

void QuickView::slotFeedNextClicked(void)
{
  goTo(NextSite, FirstItem);
}

void QuickView::slotItemNextClicked(void)
{
  goTo(SameSite, NextItem);
}

void QuickView::preview(SiteHandler *sh)
{
  if(!sh)
    return;

  mSiteHandler = sh;

  SchemeTree st;
	SchemeParser::parseFile(QDir::homeDirPath() + QString("/.krss/schemes/currentscheme"), st);
  mBackColour = st.backgroundColour();
  mTextColour = st.activeLinkColour();
  mLinkColour = st.visitedLinkColour();
  mScroller->setPaletteBackgroundColor(mBackColour);
  mScroller->viewport()->setPaletteBackgroundColor(mBackColour);
  mItem->setPaletteBackgroundColor(mBackColour);
  mItem->setHighlightedColor(mTextColour);
  mItem->setSelectedColor(mLinkColour);
  mItem->setFont(st.newsTitleFont());
  
  sh->makeSiteList(mSites, true);
  setupCommon();
  goTo(FirstSite, FirstItem);
  //mScroller->setResizePolicy(QScrollView::Manual);
}

void QuickView::goTo(GoToSite s, GoToItem i)
{
  bool site_changed = true;

  switch(s) {
    case FirstSite:
      mCurrentSite = 0;
      break;
    case NextSite:
      mCurrentSite++;
      if(mCurrentSite > mSites.count() - 1) {
        goTo(FirstSite, FirstItem);
        return;
      }
      break;
    case PreviousSite:
      if(mCurrentSite > 0)
        mCurrentSite--;
      else {
        goTo(LastSite, LastItem);
        return;
      }
      break;
    case LastSite:
      mCurrentItem = mSites.count() - 1;
      break;
    case SameSite:
      site_changed = false;
      break;
  }

  if(site_changed) {
    Site *s = mSites.at(mCurrentSite);
    if(!s) {
      if(mCurrentSite != 0) {
        goTo(FirstSite, FirstItem);
        return;
      }
    }
    RDFTree tree;
    if(!RSSParser::parseFile(s->fileNameParsed(), tree)) {
      goTo(NextSite, FirstItem);
      return;
    }

    QList<RDFItem> *items = tree.items();
    RDFItem *it = items->first();
    mItems.clear();
    while(it) {
      RDFItem ni;
      ni.setTitle(it->title());
      ni.setDescription(it->description());
      ni.setLink(it->link());
      ni.setUri(it->uri());
      mItems.append(ni);
      it = items->next();
    }

    setCaption(QString("K.R.S.S. - ") + tree.channelTitle());

    if(mItems.count() < 1) {
      goTo(NextSite, FirstItem);
      return;
    }
  }

  switch(i) {
    case FirstItem:
      mCurrentItem = 0;
      break;
    case NextItem:
      mCurrentItem++;
      if(mCurrentItem > mItems.count() - 1) {
        goTo(NextSite, FirstItem);
        return;
      }
      break;
    case PreviousItem:
      if(mCurrentItem > 0)
        mCurrentItem--;
      else {
        goTo(PreviousSite, LastItem);
        return;
      }
      break;
    case LastItem:
      mCurrentItem = mItems.count() - 1;
      break;
  }

  QValueList<RDFItem>::Iterator iter = mItems.at(mCurrentItem);

  mItem->setText((*iter).title());
  mItem->setURL((*iter).link().url());
  setupToolTip((*iter).title(), (*iter).description(), (*iter).link().url());
  //mScroller->viewport()->resize(mItem->width() * 3, mItem->height() * 3);
  switch(mScrollMethod) {
    case Vertical:
      setupVertical();
      break;
    case Horizontal:
    default:
      setupHorizontal();
      break;
  }
}

void QuickView::slotTimeout(void)
{
  mDiff++;
  switch(mScrollMethod) {
    case Vertical:
      mScroller->setContentsPos(0, mDiff);
      setupVertical(false);
      break;
    case Horizontal:
    default:
      mScroller->setContentsPos(mDiff, 0);
      setupHorizontal(false);
      break;
  }

  if(mFade && mTotal != 0) {
    double per;
    if(mDiff < mFadeDist) {
      //cout << "Fading In" << endl;
      per = (double)mDiff / (double)mFadeDist;
    }
    else if(mDiff > mTotal - mFadeDist) {
      //cout << "Fading Out" << endl;
      per = -(((double)mDiff - (double)mTotal) / mFadeDist);
    }
    else
      per = 1.0;

    //cout << "Fade Percentage: " << per * 100.0 << endl;

    int r, g, b;
    int dr, dg, db;
    dr = mBackColour.red() - mTextColour.red();
    dg = mBackColour.green() - mTextColour.green();
    db = mBackColour.blue() - mTextColour.blue();

    r = mBackColour.red() - (int)(per * (double)dr);
    g = mBackColour.green() - (int)(per * (double)dg);
    b = mBackColour.blue() - (int)(per * (double)db);

    if(r > 255) r = 255;
    if(g > 255) g = 255;
    if(b > 255) b = 255;
    if(r < 0) r = 0;
    if(g < 0) g = 0;
    if(b < 0) b = 0;

    //cout << "RGB: " << r << ", " << g << ", " << b << endl;

    mItem->setHighlightedColor(QColor(r, g, b));

    dr = mBackColour.red() - mLinkColour.red();
    dg = mBackColour.green() - mLinkColour.green();
    db = mBackColour.blue() - mLinkColour.blue();

    r = mBackColour.red() - (int)(per * (double)dr);
    g = mBackColour.green() - (int)(per * (double)dg);
    b = mBackColour.blue() - (int)(per * (double)db);

    //cout << "RGB: " << r << ", " << g << ", " << b << endl;

    if(r > 255) r = 255;
    if(g > 255) g = 255;
    if(b > 255) b = 255;
    if(r < 0) r = 0;
    if(g < 0) g = 0;
    if(b < 0) b = 0;

    mItem->setSelectedColor(QColor(r, g, b));
  }

  //mScroller->updateContents();

  if(mDiff > mTotal)
    goTo(SameSite, NextItem);
}

void QuickView::setupCommon(void)
{
  int h = mItem->fontMetrics().height();

  //Set the maximum and minimum sizes of the window
  setMaximumSize(400, h * 3);
  setMinimumSize(200, h);
  //mScroller->resize(400, h + 3);

  //Turn off the scroll bars
  mScroller->setHScrollBarMode(QScrollView::AlwaysOff);
  mScroller->setVScrollBarMode(QScrollView::AlwaysOff);
}

void QuickView::setupFading(void)
{
  mFadeDist = ((10 - mFadeSpeed) + 1) * 5;

  if(mFadeDist == 0)
    mFadeDist = 1;

  if(mTotal == 0)
    mTotal = 1;

  mFadeDist = (int)(((double)mFadeDist / 100.0) * (double)mTotal);
}

void QuickView::setupHorizontal(bool newitem)
{
  if(newitem) {
    //Move the Item to its inital position
    mScroller->setContentsPos(0, 0);
    mScroller->moveChild(mItem, mScroller->visibleWidth(), 0);

    //Setup some things to determine when we are done with this feed
    mDiff = 0;
    mTotal = mItem->fontMetrics().width(mItem->text()) + mScroller->visibleWidth();

    setupFading();
  }

  mItem->setSizePolicy(QSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding));
  //Change the height of the item
  mItem->resize(mItem->fontMetrics().width(mItem->text()), mScroller->visibleHeight() + 2);

  //40,000 is a nice safe size
  mScroller->resizeContents(40000, 40000);
  mScroller->updateContents();
}

void QuickView::setupVertical(bool newitem)
{
  mItem->resize(mScroller->visibleWidth(), mItem->height());

  if(newitem) {
    //Move the Item to its inital position
    mScroller->setContentsPos(0, 0);
    mScroller->moveChild(mItem, 0, mScroller->visibleHeight());

    //40,000 is a nice safe size
    //mScroller->resizeContents(mScroller->visibleWidth(), 40000);
    QRect rect(0, 0, mScroller->visibleWidth(), mItem->fontMetrics().height() * 100);
    QString text = mItem->text();
    QFontMetrics fm = mItem->fontMetrics();
    KWordWrap *ww = KWordWrap::formatText(fm, rect, 0, text, -1);
    if(ww) {
      QString wstr = ww->wrappedString();
      unsigned int nlines = 0;
      QTextStream t(wstr, IO_ReadOnly);
      while(!t.atEnd()) {
        QString tstr = t.readLine();
        nlines++;
      }
      mItem->resize(ww->boundingRect().width(), mItem->fontMetrics().height() * nlines);
      mItem->setText(ww->wrappedString());
      delete ww;
      ww = 0;
    }

    mDiff = 0;
    mTotal = mItem->height() + mScroller->visibleHeight();

    setupFading();
  }

  //400,000 is a nice safe size
  mScroller->resizeContents(400000, 400000);
  mScroller->updateContents();
}

void QuickView::slotLeftClickURL(const QString & url)
{
  KConfig * config = KGlobal::config();
  config->setGroup(QString("General"));
	QString mExec = config->readEntry(QString("Open URLs With"), QString("konqueror %url"));
	mExec.replace(QRegExp("%url"), url);
	
	KProcess p; // = new KProcess;
	QTextStream t(&mExec, IO_ReadOnly);
	while(!t.atEnd()) {
		QString mStr;
		t >> mStr;
		p << mStr;
	}

	p.start(KProcess::DontCare);
}

void QuickView::slotRightClickURL(const QString & url)
{
  if(url.isEmpty() || url.isNull()) {
		return;
	}
	else {
		KPopupMenu menu(this, "url_menu");
		menu.insertTitle(i18n("Open URL Using:"));
    KConfig *config = KGlobal::config();
		config->setGroup(QString("General"));
		QStringList uh_list = config->readListEntry(QString("URL Handlers"));
		QString default_handler = config->readEntry(QString("Open URLs With"), QString("konqueror %url"));
		default_handler.replace(QRegExp("%url"), url);
		menu.insertItem(default_handler);
		menu.insertSeparator();
		for(QStringList::Iterator i = uh_list.begin(); i != uh_list.end(); i++) {
			QString handler((*i).latin1());
			handler.replace(QRegExp("%url"), url);
			menu.insertItem(handler);
		}
		int ret = menu.exec();

		if(ret == -1)
			return;

		QString mExec = menu.text(ret);
		//mExec.replace(QString("%url"), url);


		KProcess p; // = new KProcess;
		QTextStream t(&mExec, IO_ReadOnly);
		while(!t.atEnd()) {
			QString mStr;
			t >> mStr;
			p << mStr;
		}

		p.start(KProcess::DontCare);
	}
}

void QuickView::slotScrollerResized(int , int , int , int )
{
  switch(mScrollMethod) {
    case Vertical:
        setupVertical(false);
      break;
    case Horizontal:
        setupHorizontal(false);
    default:
      break;
  }
}

void QuickView::slotContentsDragEntered(QDragEnterEvent* event)
{
	event->accept(KURLDrag::canDecode(event));
}

void QuickView::slotContentsDroped(QDropEvent* event)
{
	KURL::List urls;
	if(KURLDrag::decode(event, urls)) {
		KURL dest;
		dest.setPath(QDir::homeDirPath() + QString("/.krss/tmp/"));
		KIO::CopyJob * cpjob = KIO::copy(urls, dest, true);
		connect(cpjob, SIGNAL(result(KIO::Job *)), this, SLOT(slotResult(KIO::Job *)));
		connect(cpjob, SIGNAL(copyingDone(KIO::Job *, const KURL &, const KURL &, bool, bool)), this, SLOT(slotCopyingDone(KIO::Job *, const KURL &, const KURL &, bool, bool)));
	}
}

void QuickView::slotResult(KIO::Job *)
{
	emit signalDnDFinished();
}

void QuickView::slotDnDFinished(void)
{
    preview(mSiteHandler);
}

void QuickView::slotCopyingDone(KIO::Job *, const KURL & from, const KURL & to, bool, bool)
{
	QFile f(to.path());
	if(!f.exists()) return;

	RSSParser parser;
	QFile xmlFile(to.path());
#ifndef USE_INTERNAL_RDF_HANDLER
	QXmlInputSource source(xmlFile);
	QXmlSimpleReader reader;
	reader.setContentHandler(&parser);
	reader.setErrorHandler(&parser);
	bool ret = reader.parse(source);
#else
  if(!xmlFile.open(IO_ReadOnly))
    return;

  QTextStream t(&xmlFile);
  bool ret = parser.parse(t.read());
#endif

  if(!ret) return;


	Site * s = 0;

	s = new Site();

	s->setName(parser.rdfTree().channelTitle());
	s->setURL(from);
	s->setDescription(parser.rdfTree().channelDescription());
	s->setUsed(true);
	s->makeFileName();

	if(mSiteHandler) {
		mSiteHandler->addSite(s);
		mSiteHandler->writeSites();
	}

	QFile fw(s->fileNameParsed());

	if(!f.open(IO_ReadOnly)) return;
	if(!fw.open(IO_WriteOnly)) return;

	QTextStream tr(&f);
	QTextStream tw(&fw);

	tw << tr.read();

	f.close();
	fw.close();

	QDir().remove(to.path(), true);
}

void QuickView::setupToolTip(const QString & title, const QString & desc, const QString & url)
{
  QString ttip = QString("<qt><b>");

  ttip.append(title);
  ttip.append(QString("</b><br>"));

  ttip.append(desc);
  ttip.append(QString("<br>"));

  ttip.append(url);

  ttip.append(QString("</qt>"));

  mItem->setTipText(ttip);
}


